"""Top-level package for SLURM Completed Logs Utils."""

__author__ = """Jaideep Sundaram"""
__email__ = 'jai.python3@gmail.com'
__version__ = '0.1.0'

from .parser import Parser as SlurmCompletedLogsParser # noqa F401
from .record import Record as SlurmCompletedLogsRecord # noqa F401



import logging
import os

from typing import List

from .. import constants
from ..record import Record
from ..file_utils import calculate_md5, get_file_creation_date
from .dbutil import (
    create_database,
    create_provenance_table,
    create_table,
    insert_provenance_table,
    insert_record_into_table,
)


class Manager:
    """Class loading the SLURM Completed Jobs records into a SQLite3 database instance."""

    def __init__(self, **kwargs):
        """Constructor for Manager."""
        self.config = kwargs.get("config", None)
        self.config_file = kwargs.get("config_file", None)
        self.database_file = kwargs.get("database_file", None)
        self.infile = kwargs.get("infile", None)
        self.logfile = kwargs.get("logfile", None)
        self.outdir = kwargs.get("outdir", None)
        self.table_name = kwargs.get("table_name", constants.DEFAULT_TABLE_NAME)
        self.verbose = kwargs.get("verbose", constants.DEFAULT_VERBOSE)

        self.filtered_column_names = None
        self.ignore_column_lookup = {}
        self._load_ignore_columns_lookup()

        logging.info(f"Instantiated Manager in file '{os.path.abspath(__file__)}'")

    def _load_ignore_columns_lookup(self) -> None:
        ignore_columns_list = self.config.get("ignore_columns", {})
        ctr = 0
        for column in ignore_columns_list:
            ctr += 1
            self.ignore_column_lookup[column] = True

        if ctr > 0:
            logging.info(f"Loaded '{ctr}' column names into the ignore column lookup")

    def load_records(self, records: List[Record]) -> None:
        """Load the SLURM Completed Jobs log records into a SQLite3 database
        instance."""
        create_database(self.database_file)
        create_provenance_table()
        self._insert_provenance()
        # TODO: add support to check whether the table already exists
        self._create_jobs_table()
        self._load_records_into_jobs_table(records)

    def _create_jobs_table(self) -> None:
        """Create the jobs table in the SQLite database."""

        column_names = self._get_filtered_column_names()

        create_table(
            self.table_name,
            column_names,
            self.config
        )

    def _get_filtered_column_names(self) -> List[str]:
        column_names = []
        if "table_schema" not in self.config:
            raise Exception(f"'table_schema' not defined in the configuration file '{self.config_file}'")
        if "jobs" not in self.config["table_schema"]:
            raise Exception(f"'jobs' not defined in the 'table_schema' section of the configuration file '{self.config_file}'")

        column_names = self.config["table_schema"]["jobs"]

        filtered_column_names = []

        for column_name in column_names:
            if column_name in self.ignore_column_lookup:
                logging.info(f"Excluding column '{column_name}'")
                continue
            filtered_column_names.append(column_name)

        self.filtered_column_names = filtered_column_names

        return filtered_column_names

    def _load_records_into_jobs_table(self, records: List[Record]) -> None:
        """Load the records into the jobs table in the SQLite database.

        Args:
            records (List[Record]): The list of records to load into the database.
        """
        for record in records:
            record_dict = record.model_dump()
            final_record_list = []
            final_column_names = []
            for key, value in record_dict.items():
                if key in self.filtered_column_names:
                    final_column_names.append(key)
                    final_record_list.append(value)
            insert_record_into_table(self.table_name, final_column_names, final_record_list)

    def _get_normalized_name(self, column_name: str) -> str:
        """Derive a normalized column name for the column.

        This will replace all punctuation, symbols and spaces with underscores

        Args:
            column_name (str): the original column name

        Returns:
            str: the normalized column name
        """
        column_name = (
            column_name.replace(" ", "_")
            .replace("/", "_")
            .replace("\\", "_")
            .replace("(", "_")
            .replace(")", "_")
            .replace("*", "_")
            .replace("&", "_")
            .replace("@", "_")
            .replace("#", "")
            .replace("!", "_")
            .replace("?", "_")
            .replace(".", "_")
            .replace("+", "_")
            .replace("-", "_")
            .replace("~", "_")
            .replace("[", "_")
            .replace("]", "_")
            .replace("{", "_")
            .replace("}", "_")
            .replace("|", "_")
            .replace(",", "_")
            .replace(";", "_")
            .replace("'", "_")
            .replace('"', "_")
        )

        # If column_name contains two or more consecutive underscores, replace with a single underscore
        while "__" in column_name:
            column_name = column_name.replace("__", "_")

        # If column_name starts with a number, prepend with an underscore
        if column_name[0].isdigit():
            column_name = f"_{column_name}"

        return column_name

    def _insert_provenance(self) -> None:
        bytesize = os.path.getsize(self.infile)
        checksum = calculate_md5(self.infile)
        date_created = get_file_creation_date(self.infile)

        insert_provenance_table(
            bytesize,
            checksum,
            date_created,
            os.path.abspath(self.infile)
        )



"""Parse the SLURM Completed Logs file and load into a table in SQLite."""
import click
import logging
import os
import pathlib
import sys
import yaml


from rich.console import Console

from .sqlite.manager import Manager as SQLiteManager
from .parser import Parser as SlurmCompletedLogsParser
from .file_utils import check_infile_status
from .console_helper import print_yellow, print_green
from . import constants


DEFAULT_OUTDIR = os.path.join(
    constants.DEFAULT_OUTDIR_BASE,
    os.path.splitext(os.path.basename(__file__))[0],
    constants.DEFAULT_TIMESTAMP
)


error_console = Console(stderr=True, style="bold red")

console = Console()


def validate_verbose(ctx, param, value):
    """Validate the validate option.

    Args:
        ctx (Context): The click context.
        param (str): The parameter.
        value (bool): The value.

    Returns:
        bool: The value.
    """

    if value is None:
        click.secho("--verbose was not specified and therefore was set to 'True'", fg='yellow')
        return constants.DEFAULT_VERBOSE
    return value


@click.command()
@click.option(
    "--config_file",
    type=click.Path(exists=True),
    help=f"The configuration file for this project - default is '{constants.DEFAULT_CONFIG_FILE}'",
)
@click.option(
    "--database_file",
    help=f"Optional: The SQLite3 database file - default will be [infile].{constants.DEFAULT_DATABASE_FILE_EXTENSION}.",
)
@click.option(
    "--infile",
    help="Required: The input SLURM Completed Logs text file to be parsed and loaded into the SQLite3 database",
)
@click.option("--logfile", help="Optional: The log file for the execution of this program.")
@click.option(
    "--outdir",
    help=f"Optional: The default is the current working directory - default is '{DEFAULT_OUTDIR}'",
)
@click.option('--verbose', is_flag=True, help=f"Optional: Will print more info to STDOUT - default is '{constants.DEFAULT_VERBOSE}'.", callback=validate_verbose)
def main(
    config_file: str,
    database_file: str,
    infile: str,
    logfile: str,
    outdir: str,
    verbose: bool,
):
    """Parse the SLURM Completed Logs file and load into a table in SQLite."""
    error_ctr = 0

    if infile is None:
        error_console.print("--infile was not specified")
        error_ctr += 1

    if error_ctr > 0:
        click.echo(click.get_current_context().get_help())
        sys.exit(1)

    check_infile_status(infile)

    if config_file is None:
        config_file = constants.DEFAULT_CONFIG_FILE
        print_yellow(f"--config_file was not specified and therefore was set to '{config_file}'")

    check_infile_status(config_file)

    if outdir is None:
        outdir = DEFAULT_OUTDIR
        print_yellow(f"--outdir was not specified and therefore was set to '{outdir}'")

    if not os.path.exists(outdir):
        pathlib.Path(outdir).mkdir(parents=True, exist_ok=True)
        print_yellow(f"Created output directory '{outdir}'")

    if logfile is None:
        logfile = os.path.join(
            outdir, os.path.splitext(os.path.basename(__file__))[0] + ".log"
        )
        print_yellow(f"--logfile was not specified and therefore was set to '{logfile}'")

    if verbose is None:
        verbose = constants.DEFAULT_VERBOSE
        print_yellow(f"--verbose was not specified and therefore was set to '{verbose}'")

    logging.basicConfig(
        filename=logfile,
        format=constants.DEFAULT_LOGGING_FORMAT,
        level=constants.DEFAULT_LOGGING_LEVEL,
    )

    # Read the configuration from the JSON file and
    # load into dictionary.
    logging.info(f"Will load contents of config file '{config_file}'")
    config = yaml.safe_load(pathlib.Path(config_file).read_text())

    if database_file is None:
        extension = config.get(
            "database_file_extension", constants.DEFAULT_DATABASE_FILE_EXTENSION
        )
        database_file = f"{os.path.join(infile)}.{extension}"
        print_yellow(f"--database_file was not specified and therefore was set to '{database_file}'")

    manager = SQLiteManager(
        config=config,
        config_file=config_file,
        outdir=outdir,
        logfile=logfile,
        infile=infile,
        database_file=database_file,
        verbose=verbose
    )

    parser = SlurmCompletedLogsParser()

    records = parser.get_records(infile)

    manager.load_records(records)

    if verbose:
        console.print(f"The log file is '{logfile}'")
        print_green(f"Execution of '{os.path.abspath(__file__)}' completed")


if __name__ == "__main__":
    main()  # pragma: no cover



"""Module for console helper functions.

Raises:
    Exception: If msg is not defined
"""
from rich.console import Console

error_console = Console(stderr=True, style="bold red")

console = Console()


def print_red(msg: str = None) -> None:
    """Print a message in red.

    Args:
        msg (str, optional): The message to be printed in red. Defaults to None.

    Raises:
        Exception: If msg is not defined
    """
    if msg is None or msg == "":
        raise Exception("msg was not defined")
    error_console.print(msg)


def print_green(msg: str = None) -> None:
    """Print a message in green.

    Args:
        msg (str, optional): The message to be printed in red. Defaults to None.

    Raises:
        Exception: If msg is not defined
    """
    if msg is None or msg == "":
        raise Exception("msg was not defined")

    console.print(f"[bold green]{msg}[/]")


def print_yellow(msg: str = None) -> None:
    """Print a message in yellow.

    Args:
        msg (str, optional): The message to be printed in red. Defaults to None.

    Raises:
        Exception: If msg is not defined
    """
    if msg is None or msg == "":
        raise Exception("msg was not defined")

    console.print(f"[bold yellow]{msg}[/]")



"""Class for parsing the SLURM Completed Logs text files."""
import logging
import os

from .record import Record
from . import constants
from .file_utils import check_infile_status


class Parser:
    """Class for parsing the SLURM Completed Logs files."""

    def __init__(self, **kwargs):
        """Constructor for Parser"""
        self.config = kwargs.get("config", None)
        self.config_file = kwargs.get("config_file", None)
        self.logfile = kwargs.get("logfile", None)
        self.outdir = kwargs.get("outdir", None)
        self.verbose = kwargs.get("verbose", constants.DEFAULT_VERBOSE)

        self.is_parsed = False
        self.rec_ctr = 0
        self.rec_list = []

        logging.info(f"Instantiated Parser in file '{os.path.abspath(__file__)}'")

    def get_records(self, infile: str) -> bool:
        """Parse the SLURM Completed Logs text file and retrieve a list of records.

        Args:
            infile (str): The SLURM Completed Logs text file to be parsed.
        Returns:
            List(Record): The parsed records.
        """
        if self.is_parsed:
            return self.rec_list

        check_infile_status(infile)

        # Sample record:
        # JobId=24226086 UserId=root(0) GroupId=root(0) Name=2419987.AGGAGGAA-GTGCTTAC_sample_internal_status_fail.sh JobState=CANCELLED Partition=PROD TimeLimit=UNLIMITED StartTime=2024-02-26T14:35:32 EndTime=2024-02-26T14:35:32 NodeList=(null) NodeCnt=0 ProcCnt=0 WorkDir=/root ReservationName= Gres= Account=root QOS=panels WcKey= Cluster=cluster SubmitTime=2024-02-24T16:05:19 EligibleTime=unknown DerivedExitCode=0:0 ExitCode=0:0

        with open(infile) as f:
            for line in f:
                parts = line.strip().split()
                if len(parts) == 0:
                    continue
                record = Record(
                    jobid=parts[0].split('=')[1],
                    userid=parts[1].split('=')[1],
                    groupid=parts[2].split('=')[1],
                    name=parts[3].split('=')[1],
                    jobstate=parts[4].split('=')[1],
                    partition=parts[5].split('=')[1],
                    timelimit=parts[6].split('=')[1],
                    starttime=parts[7].split('=')[1],
                    endtime=parts[8].split('=')[1],
                    nodelist=parts[9].split('=')[1],
                    nodecnt=parts[10].split('=')[1],
                    proccnt=parts[11].split('=')[1],
                    workdir=parts[12].split('=')[1],
                    reservationname=parts[13].split('=')[1],
                    gres=parts[14].split('=')[1],
                    account=parts[15].split('=')[1],
                    qos=parts[16].split('=')[1],
                    wckey=parts[17].split('=')[1],
                    cluster=parts[18].split('=')[1],
                    submittime=parts[19].split('=')[1],
                    eligibletime=parts[20].split('=')[1],
                    derivedexitcode=parts[21].split('=')[1],
                    exitcode=parts[22].split('=')[1],
                )

                self.rec_list.append(record)
                self.rec_ctr += 1

        logging.info(f"Parsed '{self.rec_ctr}' records in SLURM Completed Logs file '{infile}'")

        self.is_parsed = True
        return self.rec_list



import logging
import sqlite3
import copy

from typing import Any, Dict, List
from datetime import datetime

from .. import constants


__sqlite_conn = None


def create_database(database_file: str) -> None:
    """Create connection with SQLite3 database instance."""
    # Connect to SQLite database (or create it if it doesn't exist)
    global __sqlite_conn
    __sqlite_conn = sqlite3.connect(database_file)
    logging.info(f"Established connection with SQLite3 database '{database_file}'")


def create_provenance_table() -> None:
    # Create a cursor object to interact with the database
    cursor = __sqlite_conn.cursor()

    # Create a table named 'provenance'
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS provenance (
            bytesize INTEGER NOT NULL,
            abspath TEXT NOT NULL,
            md5checksum TEXT NOT NULL,
            date_created TEXT NOT NULL
        )
    """
    )
    logging.info("Created table 'provenance'")


def create_columnmaps_table() -> None:
    # Create a cursor object to interact with the database
    cursor = __sqlite_conn.cursor()

    # Create a table named 'columnmaps'
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS columnmaps (
            column_num INTEGER NOT NULL PRIMARY KEY,
            name TEXT NOT NULL,
            norm_name TEXT NOT NULL,
            UNIQUE (name, norm_name)
        )
    """
    )
    logging.info("Created table 'columnmaps'")


def create_records_table() -> None:
    # Create a cursor object to interact with the database
    cursor = __sqlite_conn.cursor()

    # Create a table named 'records'
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS records (
            line_num INTEGER NOT NULL,
            column_num INTEGER NOT NULL,
            value TEXT NOT NULL,
            UNIQUE (line_num, column_num),
            FOREIGN KEY (column_num) REFERENCES columnmaps (column_num)
        )
    """
    )
    logging.info("Created table 'records'")


def insert_provenance_table(
    bytesize: int,
    checksum: str,
    date_created: datetime,
    infile: str
) -> None:
    """Insert record into the provenance table.

    Args:
        bytesize (int): the size of the file in bytes
        checksum (str): the MD5 checksum of the file
        date_created (datetime): the date the file was created
        infile (str): the absolute path of the file
    """
    cursor = __sqlite_conn.cursor()
    cursor.execute(
        "INSERT INTO provenance (bytesize, abspath, md5checksum, date_created) VALUES (?, ?, ?, ?)",
        (bytesize, infile, checksum, date_created),
    )
    logging.info(f"Inserted metadata into provenance table for file '{infile}'")
    __sqlite_conn.commit()


def insert_columnmaps_table(column_num: int, name: str, normalized_name: str) -> None:
    cursor = __sqlite_conn.cursor()
    cursor.execute(
        "INSERT INTO columnmaps (column_num, name, norm_name) VALUES (?, ?, ?)",
        (column_num, name, normalized_name),
    )
    logging.info(
        f"Inserted mapping into columnmaps table for original column name '{name}'"
    )
    __sqlite_conn.commit()


def insert_records_table(line_num: int, column_num: int, value: str) -> None:
    cursor = __sqlite_conn.cursor()
    cursor.execute(
        "INSERT INTO records (line_num, column_num, value) VALUES (?, ?, ?)",
        (line_num, column_num, value),
    )
    logging.info(
        f"Inserted value into records table for line '{line_num}' column '{column_num}' value '{value}'"
    )
    __sqlite_conn.commit()


def create_table(
    table_name: str,
    column_names: List[str],
    config: Dict[str, Any]
) -> None:
    """Create the table in the database.

    Args:
        table_name (str): The name of the table.
        column_names (List[str]): The list of column names.
        config (Dict[str, Any]): The configuration dictionary.
    """
    # Create a cursor object to interact with the database
    cursor = __sqlite_conn.cursor()

    columns = None
    if "table_schema" in config and table_name in config["table_schema"]:
        columns = _derive_column_datatypes_from_config(table_name, column_names, config["table_schema"][table_name])
    else:
        columns = _derive_column_datatypes(column_names)

    create_table_sql = f"CREATE TABLE IF NOT EXISTS {table_name} (" + columns + ")"

    if constants.TEST_MODE:
        print(f"Running in test mode - would have executed '{create_table_sql}'")
    else:
        cursor.execute(create_table_sql)
        logging.info("Created table 'records'")


def _derive_column_datatypes_from_config(
    table_name: str,
    column_names: List[str],
    table_datatype_lookup: Dict[str, str]
) -> str:
    """Assign the datatype to each column based on configuration.

    Args:
        table_name (str): the name of the table
        column_names (List[str]): list of column names
        table_datatype_lookup (Dict[str, str]): lookup where key is the column name and value is the datatype
    Returns:
        str: comma-separated list of column name and TEXT datatype
    """
    logging.info("Will attempt to derive the column datatypes from the configuration file")
    column_datatype = []

    for column_name in column_names:
        datatype = "TEXT"  # default datatype in case the column was not specified in the configuration file

        if column_name in table_datatype_lookup:
            datatype = table_datatype_lookup[column_name]

        column_datatype.append(f"{column_name} {datatype}")

    columns = ",\n".join(column_datatype)
    return columns


def _derive_column_datatypes(column_names: List[str]) -> str:
    """Assign TEXT datatype to all columns.

    Args:
        column_names (List[str]): list of column names
    Returns:
        str: comma-separated list of column name and TEXT datatype
    """
    column_datatype = [f"{column} TEXT" for column in column_names]
    columns = ",\n".join(column_datatype)
    return columns


def insert_record_into_table(
    table_name: str,
    column_names: List[str],
    record: List[str],
) -> None:
    """Insert the record into the table.

    Args:
        table_name (str): The name of the table.
        column_names (List[str]): The list of column names.
        record (List[str]): The list of values for the record.
    """
    cursor = __sqlite_conn.cursor()

    column_names_copy = copy.copy(column_names)
    columns = ", ".join(column_names_copy)

    values = ", ".join(record) # noqa F841

    placeholders = []

    for _ in range(len(column_names_copy)):
        placeholders.append("?")

    # placeholders.append("?")
    placeholder = ", ".join(placeholders)

    # print(f"{columns=}")
    # print(f"{placeholder=}")
    # print(f"{values=}")
    # import sys
    # sys.exit(1)
    # record_tuple = tuple(record)
    insert_sql = f"INSERT INTO {table_name} ({columns}) VALUES ({placeholder})"
    # insert_sql = f"INSERT INTO {table_name} ({columns}) VALUES ({placeholder}) ({values}, " + line_number + ")"

    if constants.TEST_MODE:
        print(f"Running in test mode - would have executed '{insert_sql}' with values {record}")
    else:
        cursor.execute(insert_sql, record)
        logging.info(f"Inserted create into {table_name}")
        __sqlite_conn.commit()



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Click>=7.0',
    "Rich",
    "pydantic",
    "PyYAML",
]

test_requirements = []

setup(
    author="Jaideep Sundaram",
    author_email='jai.python3@gmail.com',
    python_requires='>=3.10',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.10',
    ],
    description="Collection of utilities for parsing the SLURM Completed Logs file.",
    entry_points={
        'console_scripts': [
            'slurm-log-to-sqlite=slurm_completed_logs_utils.completed_log_to_sqlite:main',
            # 'insert-lines=slurm_completed_logs_utils.insert_lines:main',
            # 'make-substitutions=slurm_completed_logs_utils.make_substitutions:main',
        ],
    },
    install_requires=requirements,
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='slurm_completed_logs_utils',
    name='slurm_completed_logs_utils',
    packages=find_packages(include=['slurm_completed_logs_utils', 'slurm_completed_logs_utils.*']),
    package_data={
        "slurm_completed_logs_utils": [
            "conf/config.yaml",
            "sqlite/*",
            ]
        },
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/jai-python3/slurm-completed-logs-utils',
    version='0.4.1',
    zip_safe=False,
)



from pydantic import BaseModel


class Record(BaseModel):
    """Class for holding a record from the SLURM Completed Logs file."""
    jobid: str
    userid: str
    groupid: str
    name: str
    jobstate: str
    partition: str
    timelimit: str
    starttime: str
    endtime: str
    nodelist: str
    nodecnt: str
    proccnt: str
    workdir: str
    reservationname: str
    gres: str
    account: str
    qos: str
    wckey: str
    cluster: str
    submittime: str
    eligibletime: str
    derivedexitcode: str
    exitcode: str



import logging
import os

from datetime import datetime

DEFAULT_PROJECT = "slurm-completed-logs-utils"

DEFAULT_TIMESTAMP = str(datetime.today().strftime("%Y-%m-%d-%H%M%S"))

DEFAULT_OUTDIR_BASE = os.path.join(
    "/tmp/",
    os.getenv("USER"),
    DEFAULT_PROJECT,
)

DEFAULT_LOGGING_FORMAT = "%(levelname)s : %(asctime)s : %(pathname)s : %(lineno)d : %(message)s"

DEFAULT_LOGGING_LEVEL = logging.INFO

DEFAULT_VERBOSE = False

# Standard SQLite3 database file extension
DEFAULT_DATABASE_FILE_EXTENSION = "db"

DEFAULT_TABLE_NAME = "jobs"

TEST_MODE = False


DEFAULT_CONFIG_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "conf",
    "config.yaml",
)
